// ToHeart XNvgfR[_[
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>

//---------------------------------------------------------------------------
typedef unsigned char  u_char;
typedef unsigned long  u_long;
typedef unsigned short u_short;

//---------------------------------------------------------------------------

typedef struct {
	// src
	u_char* p;
	long    size;

	// dst
	u_char* pE;
	long    sizeEvent;

	u_char* pM;
	long    sizeMessage;

	// font
	u_char* pF;

} ST_SCN;

ST_SCN Scn;

//---------------------------------------------------------------------------
void    ScnFont();
void    ScnOpen(const char* fname);
void    ScnExt();
u_char* ScnLzs(u_char* p, long DataSize);

void    ScnSave(const char* fname);
void    ScnSaveSub(char* str, u_char* p, long size, FILE* fp, int rev);
void    ScnFree();

int     ScnMapBgm(int no);
int     ScnMapEffect(int no);

//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
	printf("decoding %s... ", argv[1]);

	ScnFont();
	ScnOpen(argv[1]);

	ScnExt();
	ScnSave(argv[1]);

	ScnFree();

	printf("done.\n");

	return 0;
}
//---------------------------------------------------------------------------
void ScnFont()
{
	char* fname = "th_scn.tbl";
	int size = 5544;

	FILE* fp = fopen(fname, "rb");

	if(fp == NULL)
	{
		fprintf(stderr, "Open Error fp %s\n", fname);
		exit(1);
	}

	Scn.pF = calloc(size, sizeof(u_char));

	if(Scn.pF == NULL)
	{
		fprintf(stderr, "Calloc pF Error\n");
		exit(1);
	}

	fread(Scn.pF, 1, size, fp);
	fclose(fp);
}
//---------------------------------------------------------------------------
void ScnOpen(const char* fname)
{
	// get filesize
	int fd = open(fname, O_RDONLY);

	if(fd == -1)
	{
		fprintf(stderr, "Open Error fd %s\n", fname);
		exit(1);
	}

	struct stat stbuf;

	if(fstat(fd, &stbuf) == -1)
	{
		fprintf(stderr, "Size Error\n");
		exit(1);
	}

	Scn.size = stbuf.st_size;
	close(fd);


	// read file
	FILE* fp = fopen(fname, "rb");

	if(fp == NULL)
	{
		fprintf(stderr, "Open Error fp %s\n", fname);
		exit(1);
	}

	Scn.p = calloc(Scn.size, sizeof(u_char));

	if(Scn.p == NULL)
	{
		fprintf(stderr, "Calloc p Error\n");
		exit(1);
	}

	fread(Scn.p, 1, Scn.size, fp);
	fclose(fp);
}
//---------------------------------------------------------------------------
void ScnExt()
{
	u_char* pE = Scn.p + ((Scn.p[1] << 8) | Scn.p[0]) * 0x10;
	u_char* pM = Scn.p + ((Scn.p[3] << 8) | Scn.p[2]) * 0x10;

	long sizeE = (pE[3] << 24) | (pE[2] << 16) | (pE[1] << 8) | pE[0];
	long sizeM = (pM[3] << 24) | (pM[2] << 16) | (pM[1] << 8) | pM[0];

	// CxgƃbZ[W
	Scn.pE = ScnLzs(pE + 4, sizeE);
	Scn.pM = ScnLzs(pM + 4, sizeM);

	Scn.sizeEvent   = sizeE;
	Scn.sizeMessage = sizeM;
}
//---------------------------------------------------------------------------
u_char* ScnLzs(u_char* p, long DataSize)
{
	u_char* q = calloc(DataSize, sizeof(u_char));

	if(q == NULL)
	{
		fprintf(stderr, "ScnLzs Calloc Error\n");
		exit(1);
	}

	u_char* pLoadBuff = p;
	u_char* pSaveBuff = q;

	u_long  Index, LIndex, Len, FlagCount;
	u_short LFlag;
	u_char  TextBuff[0x1011], Flag;

	// eLXgobt@̃NA 
	memset(TextBuff, 0, sizeof(TextBuff));

	// eLXgobt@݈̏ʒu̐ݒ
    Index = 0xfee;
	FlagCount = 0;

	while (DataSize > 0)
	{
		if (FlagCount-- > 0)
		{
			Flag <<= 1;
		}
		else
		{
			// tO̎擾 
			Flag = ~(*(pLoadBuff++));
			FlagCount = 7;
		}

		if(Flag & 0x80)
		{
			TextBuff[Index++] = *pSaveBuff++ = ~(*pLoadBuff++);
			Index &= 0x0fff;
			DataSize--;
		}
		else
		{
			// ȑOɏoꏊƒ̏̎擾(2 byte)
			LFlag = ~(*(pLoadBuff) + (*(pLoadBuff + 1) << 8));
			pLoadBuff += 2;

			//  
			Len = (LFlag & 0xf) + 3;
			DataSize -= Len;

			// ȑOoʒu
			LIndex = LFlag >> 4;

			// ȑOoʒuRs[
			while(Len-- > 0)
			{
				TextBuff[Index++] = *(pSaveBuff++) = TextBuff[LIndex++];
				LIndex &= 0x0fff;
				Index  &= 0x0fff;
			}
		}
	}

	return q;
}
//---------------------------------------------------------------------------
void ScnSave(const char* fname)
{
	char sname[13];
	strncpy(sname, fname, 13);

	char* p = strchr(sname, '.') + 1;
	p[0] = 'T';
	p[1] = 'X';
	p[2] = 'T';

	FILE* fp = fopen(sname, "wb");
	if(fp == NULL)
	{
		printf("Save Error %s\n", sname);
		exit(1);
	}

	fprintf(fp, "Filename %s\n", fname);
	fprintf(fp, "Event   Size %lx\n", Scn.sizeEvent);
	fprintf(fp, "Message Size %lx\n\n", Scn.sizeMessage);


	ScnSaveSub("Event", Scn.pE, Scn.sizeEvent, fp, 0);
	ScnSaveSub("Message", Scn.pM, Scn.sizeMessage, fp, -1);

	fclose(fp);
}
//---------------------------------------------------------------------------
void ScnSaveSub(char* str, u_char* p, long size, FILE* fp, int rev)
{
	u_char* c = p;
	int cnt = ((c[1] << 8) | c[0]) + 1 + rev;

	fprintf(fp, "%s %02x \n\n", str, cnt);
	c += 2;

/*
	for(int i=0; i<size; i++)
	{
		fprintf(fp, "%02x ", c[i]);
	}
	fprintf(fp, "\n\n");
*/
	for(int i=0; i<cnt; i++)
	{
//		fprintf(fp, "%04x - %02x %04x\n", c - p, i, ((c[1] << 8) | c[0]));
		fprintf(fp, "%02x %04x\n", i, ((c[1] << 8) | c[0]));
		c += 2;
	}
	fprintf(fp, "\n");


	while(c < p + size)
	{
		fprintf(fp, "%04x - %02x ", c - p, c[0]);
//		fprintf(fp, "%04x ", c - p);

		// tHgR[h
		if(c[0] < 0x20)
		{
			fprintf(fp, "\"");

			do
			{
				int code = (c[0] << 8) + c[1];
				c += 2;

				// 󔒈ȊOi0x00j͏݂܂
				if(code != 0)
				{
					fprintf(fp, "%c%c", Scn.pF[code * 2], Scn.pF[code * 2 + 1]);
				}

			} while(c[0] < 0x20);

			fprintf(fp, "\"\n");
			continue;
		}

		switch(c[0])
		{
		// ViII
		case 0x20:
			fprintf(fp, "end\n");
			c++;
			break;

		case 0x21:
			if(c[1] == 0x03)
			{
				// ^Cg?
				fprintf(fp, "title\n");
				c += 2;
			}
			else if(c[1] == 0x08)
			{
				// iViIIՎgpj
				fprintf(fp, "skip 21 1\n");
				c += 6;
			}
			else
			{
				// TODO mglvns肱ڂ
				fprintf(fp, "skip 21 2\n");
				c += 3;
			}
			break;

		// I
		case 0x24:
			// I bZ[W
			fprintf(fp, "sel %x %x", c[2], c[1]);

			// bZ[W ItZbg
			for(int i=0; i<c[2]; i++)
			{
				fprintf(fp, " %x %x", c[3 + i*2], c[4 + i*2]);
			}
			fprintf(fp, "\n");

			// TODO siŏWvɐݒj
			c = c + 3 + c[2] * 2 + c[4+0*2];
			break;

		// Wv
		case 0x28:
			fprintf(fp, "jmp %x %x\n", c[1]*256 + c[2], c[3]);
			c += 4;
			break;

		// ubNWv
		case 0x29:
			fprintf(fp, "jmp_block %x\n", c[1]);
			c += 2;
			break;

		// ϑI
		case 0x2b:
			fprintf(fp, "sel_opt %x %x\n", c[1], c[2]);
			c += 3;
			break;

		// PUSH_A
		case 0x2d:
			fprintf(fp, "push_a %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// RETURN_A
		case 0x2e:
			fprintf(fp, "ret_a\n");
			c++;
			break;

		// PUSH_B
		case 0x2f:
			fprintf(fp, "push_b %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// RETURN_B
		case 0x30:
			fprintf(fp, "ret_b\n");
			c++;
			break;

		// tݒ
		case 0x32:
			fprintf(fp, "cal %x\n", c[1]);
			c += 2;
			break;

		// ݒ
		case 0x33:
			fprintf(fp, "time %x\n", c[1]);
			c += 2;
			break;

		// tOZ
		case 0x34:
			fprintf(fp, "flg_add %x %x\n", c[1], c[2]);
			c += 3;
			break;

		// tOZ
		case 0x35:
			fprintf(fp, "flg_sub %x %x\n", c[1], c[2]);
			c += 3;
			break;

		// wi[h
		case 0x3c:
			fprintf(fp, "bg %x\n", c[1]);
			c += 2;
			break;

		// LN^
		case 0x3d:
			fprintf(fp, "chr_clear %x\n", c[1]);
			c += 2;
			break;

		// VISUAL[h
		case 0x42:
			fprintf(fp, "vis_load %x\n", c[1]);
			c += 2;
			break;

		// LN^\
		case 0x43:
			fprintf(fp, "chr_load %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// \
		case 0x45:
			fprintf(fp, "disp %x\n", ScnMapEffect(c[1]));
			c += 2;
			break;

		case 0x47:
			if(c[1] == 1)
			{
				fprintf(fp, "sepia %x\n", c[1]);
			}
			else
			{
				fprintf(fp, "skip 47\n");
			}
			c += 2;
			break;

		// ubNWv
		case 0x55:
			fprintf(fp, "jmp_block %x\n", c[1]);
			c += 2;
			break;

		// if==
		case 0x56:
			fprintf(fp, "if== %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// if!=
		case 0x57:
			fprintf(fp, "if!= %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// if>
		case 0x58:
			fprintf(fp, "if> %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// if<
		case 0x59:
			fprintf(fp, "if< %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// if>=
		case 0x5a:
			fprintf(fp, "if>= %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// if<=
		case 0x5b:
			fprintf(fp, "if<= %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// if_bit_on
		case 0x5e:
			fprintf(fp, "if_bit_on %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// if_bit_off
		case 0x5f:
			fprintf(fp, "if_bit_off %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// flag=
		case 0x61:
			fprintf(fp, "flag= %x %x\n", c[1], c[2]);
			c += 3;
			break;

		// flag+=
		case 0x62:
			fprintf(fp, "flag+= %x %x\n", c[1], c[2]);
			c += 3;
			break;

		// flag_bit
		case 0x63:
			fprintf(fp, "flag_bit %x %x %x\n", c[1], c[2], c[3]);
			c += 4;
			break;

		// bZ[W\NA
		case 0x69:
			fprintf(fp, "msg_after_clear %x\n", c[1]);
			c += 2;
			break;

		// bZ[W\
		case 0x6a:
			fprintf(fp, "msg %x\n", c[1]);
			c += 2;
			break;

		// ʂ𖾂邭
		case 0x73:
			fprintf(fp, "light\n");
			c += 2;
			break;

		// whiteOut
		case 0x77:
			fprintf(fp, "white_out\n");
			c += 2;
			break;

		// WhiteIn
		case 0x78:
			fprintf(fp, "white_in\n");
			c += 2;
			break;

		// BGM Jn
		case 0x80:
			fprintf(fp, "bgm_start %x\n", ScnMapBgm(c[1]));
			c += 2;
			break;

		// BGM tF[h
		case 0x81:
			fprintf(fp, "bgm_fade\n");
			c++;
			break;

		// BGM ~
		case 0x82:
			fprintf(fp, "bgm_stop\n");
			c++;
			break;

		// BGM tF[hI҂
		case 0x85:
			fprintf(fp, "bgm_wait_fade\n");
			c++;
			break;

		// BGM Jn(fadein)
		case 0x87:
			fprintf(fp, "bgm_start2 %x\n", ScnMapBgm(c[1]));
			c += 2;
			break;

		// BGM NEXT Jn
		case 0x88:
			fprintf(fp, "bgm_next %x\n", ScnMapBgm(c[1]));
			c += 2;
			break;

		// Q[I
		case 0x94:
			fprintf(fp, "game_end\n");
			c++;
			break;

		// GfBOJn
		case 0x95:
			fprintf(fp, "ending\n");
			c += 2;
			break;

		// PCM [h
		case 0xa0:
			fprintf(fp, "pcm_load %x\n", c[1]);
			c += 2;
			break;

		// PCM ~
		case 0xa1:
			fprintf(fp, "pcm_stop\n");
			c++;
			break;

		// PCM Đ
		case 0xa2:
			fprintf(fp, "pcm_start %x\n", c[1]);
			c += 3;
			break;

		// PCM ~҂
		case 0xa3:
			fprintf(fp, "pcm_wait\n");
			c++;
			break;

		// ubNI
		case 0xaf:
			fprintf(fp, "block_end\n");
			c++;
			break;

		// s
		case 0xb0:
			fprintf(fp, "cr\n");
			c++;
			break;

		// L[NbN҂
		case 0xb2:
			fprintf(fp, "key\n");
			c++;
			break;

		// y[W
		case 0xb3:
			fprintf(fp, "page\n");
			c++;
			break;

		// `摬xύX(%d*10ms)
		case 0xb6:
			fprintf(fp, "wait_chr %x\n", c[1]);
			c += 2;
			break;

		// ԑ҂(%d*10ms)
		case 0xb7:
			fprintf(fp, "wait_time %x\n", c[1]);
			c += 2;
			break;

		// ʒuύX(x=%d*8)
		case 0xb9:
			fprintf(fp, "chr_xy %x\n", c[1]);
			c += 2;
			break;

		// tbVGtFNg
		case 0xbb:
			fprintf(fp, "flash\n");
			c++;
			break;

		// uǂvʐU
		case 0xbc:
			fprintf(fp, "vibrate\n");
			c += 2;
			break;

		// wi[h2
		case 0xbd:
			fprintf(fp, "bg2 %x %x %x\n", c[1], ScnMapEffect(c[2]), ScnMapEffect(c[3]));
			c += 4;
			break;

		// rWAV[摜[h
		case 0xbf:
			fprintf(fp, "vis_scene %x %x %x\n", c[1], ScnMapEffect(c[2]), ScnMapEffect(c[3]));
			c += 4;
			break;

		// HV[摜[h
		case 0xc0:
			fprintf(fp, "vis_h %x %x %x\n", c[1], ScnMapEffect(c[2]), ScnMapEffect(c[3]));
			c += 4;
			break;

		// LN^ύX
		case 0xc1:
			fprintf(fp, "chr_change %x %x\n", c[1], c[2]<<8|c[3]);
			c += 4;
			break;

		// LN^\
		case 0xc2:
			fprintf(fp, "chr_disp %x %x\n", c[1], c[2]<<8|c[3]);
			c += 4;
			break;

		// SL\
		case 0xc3:
			fprintf(fp, "chr_clear %x %x\n", c[1], c[2]<<8|c[3]);
			c += 4;
			break;

		// witʃ[h
		// witLύX(%s, C%02x%02x.LF2, S%02d?.LF2, %02x, %02x)
		case 0xc4:
			fprintf(fp, "bg_chr_load %x %x %x %x %x %x\n", c[1], c[2], c[3], c[4], ScnMapEffect(c[5]), ScnMapEffect(c[6]));
			c += 7;
			break;

		// 3L\
		case 0xc6:
			fprintf(fp, "chr_3 %x %x %x %x %x %x %x %x %x\n", c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9]);
			c += 10;
			break;

		// lO
		case 0xd1:
			fprintf(fp, "\"\"\n");
			c++;
			break;

		case 0xd2:
			fprintf(fp, "\"c\"\n");
			c++;
			break;

		case 0xd8:
			fprintf(fp, "\"_\"\n");
			c++;
			break;

		case 0xd9:
			fprintf(fp, "\"V\"\n");
			c++;
			break;

		case 0xdf:
			fprintf(fp, "\"q\"\n");
			c++;
			break;

		case 0xe0:
			fprintf(fp, "\"\"\n");
			c++;
			break;

		case 0xe6:
			fprintf(fp, "\"\"\n");
			c++;
			break;

		case 0xe7:
			fprintf(fp, "\"\"\n");
			c++;
			break;

		case 0xe8:
			fprintf(fp, "\"\"\n");
			c++;
			break;

		case 0xe9:
			fprintf(fp, "\"\"\n");
			c++;
			break;

		case 0xed:
			fprintf(fp, "\"q\"\n");
			c++;
			break;

		case 0xee:
			fprintf(fp, "\"\"\n");
			c++;
			break;

		case 0xef:
			fprintf(fp, "\"\"\n");
			c++;
			break;

		case 0xf0:
			fprintf(fp, "\"L\"\n");
			c++;
			break;

		// tXV J_[L
		case 0xf5:
			fprintf(fp, "cal_update %x\n", c[1]);
			c += 2;
			break;

		// GtFNgJA
		case 0xf7:
			if(c[1] == 0x1)
			{
				fprintf(fp, "rein\n");
			}
			else if(c[1] == 0x4)
			{
				fprintf(fp, "sakura\n");
			}
			else
			{
				fprintf(fp, "skip f7\n");
			}
			c += 2;
			break;

		// I
		case 0xff:
			fprintf(fp, "event_end\n\n");
			c++;
			break;


		// TODO mglvns肱ڂ
		case 0x25:
		case 0x3f:
			fprintf(fp, "ignore\n");
			c++;
			break;

		//  1oCg
		case 0x26:
		case 0x27:		// 1ȎIɖ߂}[N?
		case 0x2c:		// IOʒu
		case 0x50:
		case 0x6b:
		case 0x70:
		case 0x71:
		case 0x74:
		case 0x75:
		case 0x76:
		case 0x79:
		case 0x7a:
		case 0xa6:
		case 0xa8:
		case 0xb8:
		case 0xbe:
		case 0xc7:
		case 0xc9:

		case 0xd3:		// lO̗]
		case 0xd4:
		case 0xd5:
		case 0xd6:
		case 0xda:
		case 0xdb:
		case 0xdc:
		case 0xdd:
		case 0xe1:
		case 0xe2:
		case 0xe3:
		case 0xe4:
		case 0xea:
		case 0xeb:
		case 0xf1:
		case 0xf2:
			fprintf(fp, "skip1\n");
			c++;
			break;

		//  2oCg
		case 0x23:
		case 0x40:
		case 0x6c:
		case 0x72:
		case 0xb4:
		case 0xcf:
		case 0xf8:
			fprintf(fp, "skip2\n");
			c += 2;
			break;

		//  3oCg
		case 0x44:
		case 0xa7:
		case 0xba:
		case 0xcc:
			fprintf(fp, "skip3\n");
			c += 3;
			break;

		default:
			fprintf(fp, "error\n");
			c++;
			break;
		}
	}
}
//---------------------------------------------------------------------------
void ScnFree()
{
	free(Scn.pE);
	free(Scn.pM);
}
//---------------------------------------------------------------------------
int ScnMapBgm(int no)
{
	switch(no)
	{
	case 27: return 31;
	case 28: return 32;
	case 29: return 33;
	case 30: return 34;
	case 31: return 35;
	case 32: return 29; /* ??? */
	case 33: return 30;

	default:
		return no+2;
	}
}
//---------------------------------------------------------------------------
typedef enum {
	LVNS_EFFECT_NONE,
	LVNS_EFFECT_NO_EFFECT,         /* ȂɂȂ */
	LVNS_EFFECT_NORMAL,            /* P\ */
	LVNS_EFFECT_FADE_MASK,         /* }XNtF[h */
	LVNS_EFFECT_WIPE_TTOB,         /* ォ烏Cv */
	LVNS_EFFECT_WIPE_LTOR,         /* 烏Cv */
	LVNS_EFFECT_FADE_PALETTE,      /* Palette tF[h */
	LVNS_EFFECT_WIPE_MASK_LTOR,    /* 烏Cv(}XN) */
	LVNS_EFFECT_FADE_SQUARE,       /* Ђ`tF[h */
	LVNS_EFFECT_WIPE_SQUARE_LTOR,  /* Ђ`Cv */
	LVNS_EFFECT_SLIDE_LTOR,        /* cXCh */
	LVNS_EFFECT_SLANTTILE,         /* ΂ꉡUCNXCh */
	LVNS_EFFECT_GURUGURU,          /* 邮 */
	LVNS_EFFECT_VERTCOMPOSITION,   /* cł݂͂ */
	LVNS_EFFECT_CIRCLE_SHRINK,     /* Sɂ */
	LVNS_EFFECT_LEFT_SCROLL,       /* ɃXN[ */
	LVNS_EFFECT_TOP_SCROLL,        /* ɃXN[ */
	LVNS_EFFECT_RAND_RASTER,       /* _s\ */
	LVNS_EFFECT_BLOOD,             /* 炾   */
} LvnsEffectType;                  /* GtFNg */

int ScnMapEffect(int no)
{
	static int toheart_effect[] = {
		LVNS_EFFECT_FADE_PALETTE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_FADE_MASK,
		LVNS_EFFECT_FADE_SQUARE,
		LVNS_EFFECT_WIPE_SQUARE_LTOR,
		LVNS_EFFECT_FADE_MASK,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_WIPE_MASK_LTOR,
		LVNS_EFFECT_SLIDE_LTOR,

		LVNS_EFFECT_WIPE_LTOR,
		LVNS_EFFECT_NO_EFFECT,
		LVNS_EFFECT_FADE_MASK,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_TOP_SCROLL,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,

		LVNS_EFFECT_BLOOD,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,
		LVNS_EFFECT_NONE,

		LVNS_EFFECT_NONE,
		LVNS_EFFECT_FADE_PALETTE
	};

	if(no == 255)
	{
		return LVNS_EFFECT_NONE;
	}

	return toheart_effect[no];
}
